"use strict";

exports.__esModule = true;
exports.StateType = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _lodash = require("lodash.defaults");

var _lodash2 = _interopRequireDefault(_lodash);

var _lodash3 = require("lodash.isarray");

var _lodash4 = _interopRequireDefault(_lodash3);

var _ArrayProperty = require("./ArrayProperty");

var _ArrayProperty2 = _interopRequireDefault(_ArrayProperty);

var _CollectionProperty = require("./collection/CollectionProperty");

var _CollectionProperty2 = _interopRequireDefault(_CollectionProperty);

var _IndexedProperty = require("./IndexedProperty");

var _IndexedProperty2 = _interopRequireDefault(_IndexedProperty);

var _KeyedProperty = require("./KeyedProperty");

var _KeyedProperty2 = _interopRequireDefault(_KeyedProperty);

var _MapProperty = require("./MapProperty");

var _MapProperty2 = _interopRequireDefault(_MapProperty);

var _PropertyFactoryShader = require("./PropertyFactoryShader");

var _PropertyFactoryShader2 = _interopRequireDefault(_PropertyFactoryShader);

var _PrimitiveProperty = require("./PrimitiveProperty");

var _PrimitiveProperty2 = _interopRequireDefault(_PrimitiveProperty);

var _Property = require("./Property");

var _Property2 = _interopRequireDefault(_Property);

var _Shader = require("./Shader");

var _Shader2 = _interopRequireDefault(_Shader);

var _Shadow = require("./Shadow");

var _Shadow2 = _interopRequireDefault(_Shadow);

var _ShadowImpl = require("./ShadowImpl");

var _ShadowImpl2 = _interopRequireDefault(_ShadowImpl);

var _akutils = require("akutils");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/*
	Todo:
		- add managed property type support to factory shader
		- attach shader to constructor so shared. Change will require passing property into each call
		  so may not be worth effort. Would need to get rid of KeyedProperty.addProperty() methods.
*/

var StateType = exports.StateType = function () {
	function StateType(PropertyClass) {
		_classCallCheck(this, StateType);

		this._PropertyClass = PropertyClass;
		this._properties = {};
		this._elementType = null;

		// setup default values
		this._autoshadow = true;
		this._defaults = undefined;
		this._implementationClass = null;
		this._initialState = undefined;
		this._readonly = false;
		this._shadowClass = null;
	}

	_createClass(StateType, [{
		key: "addProperty",
		value: function addProperty(name, type) {
			var _this = this;

			(0, _akutils.assert)(function (a) {
				a.is(isKeyedPrototype(_this._PropertyClass), "PropertyClass must be a subclass of KeyedProperty");
			});

			this._properties[name] = type;
		}

		/*
  	Configures
  */

	}, {
		key: "configureShader",
		value: function configureShader(shader) {
			shader.setAutoshadow(this._autoshadow);

			this._setupShader(shader);
		}
	}, {
		key: "createProperty",
		value: function createProperty() {
			return new this._PropertyClass(this._initialState, this._autoshadow, this._readonly);
		}
	}, {
		key: "default",
		value: function _default(state) {
			this._defaults = state;

			return this;
		}

		/*
  	Returns a PropertyFactoryShader that will create property instances as configured by this StateType.
  */

	}, {
		key: "factory",
		value: function factory(parentProperty) {
			var shader = new _PropertyFactoryShader2.default(this._PropertyClass, parentProperty, this._initialState, this._autoshadow, this._readonly || parentProperty.readonly, this._automount);

			this._setupShader(shader);

			return shader;
		}
	}, {
		key: "getTypeName",
		value: function getTypeName() {
			return this.ProperClass.__fluxTypeName__;
		}
	}, {
		key: "implementationClass",
		value: function implementationClass(cls) {
			this._implementationClass = cls;

			return this;
		}
	}, {
		key: "initialState",
		value: function initialState(state) {
			this._initialState = state;

			return this;
		}
	}, {
		key: "isComplex",
		value: function isComplex() {
			return Object.keys(this._properties).length || this._managedPropertyType || this._elementType;
		}
	}, {
		key: "properties",
		value: function properties(propTypes) {
			var _this2 = this;

			(0, _akutils.assert)(function (a) {
				a.is(isKeyedPrototype(_this2._PropertyClass), "PropertyClass must subclass KeyedProperty or implement supportsKeyedChildProperties()");
			});

			for (var name in propTypes) {
				this.addProperty(name, propTypes[name]);
			}

			return this;
		}
	}, {
		key: "setElementType",
		value: function setElementType(elementType) {
			var _this3 = this;

			(0, _akutils.assert)(function (a) {
				return a.is(isIndexedPrototype(_this3._PropertyClass), "PropertyClass must be a subclass of IndexedProperty");
			});

			this._elementType = elementType;

			return this;
		}
	}, {
		key: "shader",
		value: function shader(property) {
			var shader = new _Shader2.default(property, property.autoShadow);

			this._setupShader(shader);

			return shader;
		}
	}, {
		key: "shadowClass",
		value: function shadowClass(cls) {
			this._shadowClass = cls;

			return this;
		}
	}, {
		key: "typeName",
		value: function typeName(name) {
			this._PropertyClass.__fluxTypeName__ = name;
		}
	}, {
		key: "_setupShader",
		value: function _setupShader(shader) {
			if (this._elementType) {
				var eltType = this._elementType;

				if (eltType.isComplex && eltType.isComplex()) {
					shader.setElementStateType(eltType);
				} else if (eltType.PropertyClass) {
					shader.setElementClass(eltType.PropertyClass, eltType.defaults, eltType.autoShadow, eltType.readonly);
				} else {
					shader.setElementClass(eltType);
				}
			} else if (this._managedPropertyType) {
				// will call the setElementType() method after property created - will need to add
				// functionality to the factory shader

				// ADD setManagedPropertyType() TO FACTORY - requires changes to Collection first

				shader.setManagedPropertyType(this._managedPropertyType);
			} else {
				var _eltType = void 0;

				for (var name in this._properties) {
					_eltType = this._properties[name];

					if (_eltType.isComplex()) {
						// complex definition so need to pass entire definition to shader so can handle recusively
						shader.addStateType(name, _eltType);
					} else {
						shader.addPropertyClass(name, _eltType._PropertyClass, _eltType._initialState, _eltType._autoshadow, _eltType._readonly);
					}
				}
			}
		}
	}, {
		key: "autoshadow",
		get: function get() {
			this._autoshadow = true;

			return this;
		}
	}, {
		key: "autoshadowOff",
		get: function get() {
			this._autoshadow = false;

			return this;
		}
	}, {
		key: "readonly",
		get: function get() {
			this._readonly = true;

			return this;
		}
	}, {
		key: "readonlyOff",
		get: function get() {
			this._readonly = false;

			return this;
		}
	}], [{
		key: "computeInitialState",
		value: function computeInitialState(property, state) {
			var proto = Object.getPrototypeOf(property);
			var stateSpec = proto.constructor.stateSpec;
			var propType, propState;

			if (!stateSpec) {
				return state;
			}

			if (state === undefined) {
				state = stateSpec._initialState;
			}

			if (!(0, _akutils.isObject)(state)) {
				return state;
			}

			var propSpecs = stateSpec._properties || stateSpec;

			for (var name in propSpecs) {
				propType = propSpecs[name];
				propState = state[name];

				state[name] = propState === undefined ? propType._initialState : propState;
			}

			return state;
		}
	}, {
		key: "implementationClassForProperty",
		value: function implementationClassForProperty(property) {
			var defaultClass = arguments.length <= 1 || arguments[1] === undefined ? _ShadowImpl2.default : arguments[1];

			var proto = Object.getPrototypeOf(property);
			var stateSpec = proto.constructor.stateSpec;

			if (stateSpec && stateSpec._implementationClass) {
				return stateSpec._implementationClass;
			} else {
				return defaultClass;
			}
		}
	}, {
		key: "initialStateWithDefaults",
		value: function initialStateWithDefaults(property, state) {
			var proto = Object.getPrototypeOf(property);
			var stateSpec = proto.constructor.stateSpec;
			var arrayType = (0, _lodash4.default)(state);
			var defaultState, propType, propState;

			if (!stateSpec || !(0, _akutils.isObject)(state) && !arrayType) {
				return state;
			}

			var propSpecs = stateSpec._properties || stateSpec;

			if ((0, _akutils.isObject)(stateSpec._defaults)) {
				state = (0, _lodash2.default)(arrayType ? [] : {}, state, stateSpec._defaults);
			}

			for (var name in propSpecs) {
				propType = propSpecs[name];

				if (propType && propType._defaults !== undefined) {
					defaultState = defaultState || (arrayType ? [] : {});
					defaultState[name] = propType._defaults;
				}
			}

			if (defaultState) {
				state = (0, _lodash2.default)(arrayType ? [] : {}, state, defaultState);
			}

			return state;
		}
	}, {
		key: "shaderFromSpec",
		value: function shaderFromSpec(property, stateSpec) {
			var shader = new _Shader2.default(property, property.autoShadow);
			var eltType;

			for (var name in stateSpec) {
				eltType = stateSpec[name];

				if (eltType.isComplex()) {
					// complex definition so need to pass entire definition to shader so can handle recusively
					shader.addStateType(name, eltType);
				} else {
					shader.addPropertyClass(name, eltType._PropertyClass, eltType._initialState, eltType._autoshadow, eltType._readonly);
				}
			}

			return shader;
		}
	}, {
		key: "shadowClassForProperty",
		value: function shadowClassForProperty(property) {
			var defaultClass = arguments.length <= 1 || arguments[1] === undefined ? _Shadow2.default : arguments[1];

			var proto = Object.getPrototypeOf(property);
			var stateSpec = proto.constructor.stateSpec;

			if (stateSpec && stateSpec._shadowClass) {
				return stateSpec._shadowClass;
			} else {
				return defaultClass;
			}
		}
	}]);

	return StateType;
}();

function isIndexedPrototype(obj) {
	return _IndexedProperty2.default === obj || _IndexedProperty2.default.isPrototypeOf(obj);
}

function isKeyedPrototype(obj) {
	return _KeyedProperty2.default === obj || _KeyedProperty2.default.isPrototypeOf(obj) || _Property2.default.isPrototypeOf(obj) && obj.supportsKeyedChildProperties && obj.supportsKeyedChildProperties();
}

exports.default = {
	get Primitive() {
		return new StateType(_PrimitiveProperty2.default);
	},

	get Array() {
		var type = new StateType(_ArrayProperty2.default);

		return type;
	},

	arrayOf: function arrayOf(elementStateType) {
		var type = new StateType(_ArrayProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},


	get Indexed() {
		var type = new StateType(_IndexedProperty2.default);

		return type;
	},

	indexedOf: function indexedOf(elementStateType) {
		var type = new StateType(_IndexedProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},


	// CollectionProperty needs refactoring to support
	// collectionOf(elementStateType) {
	// 	const type = new StateType(CollectionProperty);

	// 	// will call the setElementType() method after property created - will need to add
	// 	// functionality to the factory shader
	// 	type.setManagedPropertyType(elementStateType);

	// 	return type;
	// },

	keyed: function keyed() {
		var defn = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

		var type = new StateType(_KeyedProperty2.default);
		var propType;

		for (var key in defn) {
			propType = defn[key];

			type.addProperty(key, defn[key]);
		}

		return type;
	},
	map: function map() {
		var defn = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

		var type = new StateType(_MapProperty2.default);
		var propType;

		for (var key in defn) {
			propType = defn[key];

			type.addProperty(key, defn[key]);
		}

		return type;
	},
	mapOf: function mapOf(elementStateType) {
		var type = new StateType(_MapProperty2.default);

		type.setElementType(elementStateType);

		return type;
	},
	property: function property(PropertyClass) {
		if (PropertyClass.stateSpec) {
			return PropertyClass.stateSpec;
		}

		return new StateType(PropertyClass);
	}
};