"use strict";

exports.__esModule = true;
exports.default = exports.AllOp = exports.UpdateOp = exports.FindOp = exports.FetchOp = exports.DestroyOp = exports.CreateOp = exports.ErrorEvent = exports.ChangeEvent = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _get = function get(object, property, receiver) { if (object === null) object = Function.prototype; var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { return get(parent, property, receiver); } } else if ("value" in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } };

var _componentEmitter = require("component-emitter");

var _componentEmitter2 = _interopRequireDefault(_componentEmitter);

var _invariant = require("invariant");

var _invariant2 = _interopRequireDefault(_invariant);

var _ArrayProperty = require("../ArrayProperty");

var _ArrayProperty2 = _interopRequireDefault(_ArrayProperty);

var _KeyedProperty2 = require("../KeyedProperty");

var _KeyedProperty3 = _interopRequireDefault(_KeyedProperty2);

var _MapProperty = require("../MapProperty");

var _MapProperty2 = _interopRequireDefault(_MapProperty);

var _PrimitiveProperty = require("../PrimitiveProperty");

var _PrimitiveProperty2 = _interopRequireDefault(_PrimitiveProperty);

var _Store = require("../Store");

var _Store2 = _interopRequireDefault(_Store);

var _akutils = require("akutils");

var _CollectionShadow = require("./CollectionShadow");

var _CollectionShadow2 = _interopRequireDefault(_CollectionShadow);

var _ModelProperty = require("./ModelProperty");

var _ModelProperty2 = _interopRequireDefault(_ModelProperty);

var _debug = require("../debug");

var _debug2 = _interopRequireDefault(_debug);

var _CollectionOptions = require("./CollectionOptions");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var debug = (0, _debug2.default)(_debug.CollectionPropertyKey);

var _endpoint = "_endpoint";
//const _fetching = '_fetching';
var _id2cid = "_id2cid";
var _idName = "idName";
var _lastPageSize = 'lastPageSize';
var _limit = '_limit';
var _models = "_models";
var _nextOffset = 'nextOffset';
var _paging = '_paging';
var _synced = "synced";

var _offlineState = Symbol('offlineState');
var _fetching = Symbol('fetching');
var _middleware = Symbol("middleware");

/*
	Event emitted on collection changes.
*/
var ChangeEvent = exports.ChangeEvent = "change";
/*
	Event emitted on error during an operation.
*/
var ErrorEvent = exports.ErrorEvent = "error";

/*
	Middleware operation for create requests.
*/
var CreateOp = exports.CreateOp = "create";
/*
	Middleware operation for destroy requests.
*/
var DestroyOp = exports.DestroyOp = "destroy";
/*
	Middleware operation for fetch requests.
*/
var FetchOp = exports.FetchOp = "fetch";
/*
	Middleware operation for find requests.
*/
var FindOp = exports.FindOp = "find";
/*
	Middleware operation for update requests.
*/
var UpdateOp = exports.UpdateOp = "update";
/*
	All middleware operations.
*/
var AllOp = exports.AllOp = [CreateOp, DestroyOp, FetchOp, FindOp, UpdateOp];

/*
	Temporary convenient function until the '()' transition is complete
*/
function shadow$(shadow) {
	return typeof shadow.$ === 'function' ? shadow.$() : shadow.$;
}

/*
	Rework API:
		- same ctor as all other properties
		- separate method to setup element shader:
			- setElementShader(shader/factoryShader)
			- setElementType(type, initState, autoshadow, readonly)
		- method to get element shader (shader.modelsShader.element)
*/

var CollectionProperty = function (_KeyedProperty) {
	_inherits(CollectionProperty, _KeyedProperty);

	function CollectionProperty() {
		var MemberPropertyClass = arguments.length <= 0 || arguments[0] === undefined ? _MapProperty2.default : arguments[0];

		var _this$_middleware;

		var autoShadow = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];
		var readonly = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];

		_classCallCheck(this, CollectionProperty);

		var _this = _possibleConstructorReturn(this, Object.getPrototypeOf(CollectionProperty).call(this, {}, false));

		_this[_middleware] = (_this$_middleware = {}, _defineProperty(_this$_middleware, CreateOp, []), _defineProperty(_this$_middleware, DestroyOp, []), _defineProperty(_this$_middleware, FetchOp, []), _defineProperty(_this$_middleware, FindOp, []), _defineProperty(_this$_middleware, UpdateOp, []), _this$_middleware);

		// keep isFetching as an instance variable because transient data and gives immediate
		// feedback to prevent concurrent fetches
		_this[_fetching] = false;

		_this.addProperty(_idName, new _PrimitiveProperty2.default("id", true));
		_this.addProperty(_id2cid, new _MapProperty2.default({}, true));
		_this.addProperty(_synced, new _PrimitiveProperty2.default(false, false, true));

		// pagingTime instance variable used for ensuring overlapping paging requests do not mess up offset
		// this can happen when a paging request is in progress when limit is changed
		_this.pagingTime = null;
		_this.addProperty(_lastPageSize, new _PrimitiveProperty2.default(null, false, true));
		_this.addProperty(_limit, new _PrimitiveProperty2.default(50, false, true));
		_this.addProperty(_nextOffset, new _PrimitiveProperty2.default(0, false, true));
		_this.addProperty(_paging, new _PrimitiveProperty2.default(false, false, true));

		// '_models' property contains ModelProperty objects which in turn keep their model state in
		// the 'data' whose type is specified by the 'MemberPropertyClass' parameter.
		var models = new _MapProperty2.default();
		var modelsShader = models.shader();

		_this.addProperty(_models, models);
		modelsShader.setElementClass(_ModelProperty2.default, {}, true, false);
		modelsShader.elementShader.addPropertyClass("data", MemberPropertyClass, {}, autoShadow, readonly);
		return _this;
	}

	_createClass(CollectionProperty, [{
		key: "onPropertyDidUpdate",
		value: function onPropertyDidUpdate() {
			_get(Object.getPrototypeOf(CollectionProperty.prototype), "onPropertyWillUpdate", this).call(this);

			this.emit(ChangeEvent, this._, this);
		}

		//------------------------------------------------------------------------------------------------------
		// Offline data support API
		//------------------------------------------------------------------------------------------------------

		/*
  	Experimental feature
  		Gets an object capable of providing a persisted version of this collection. The object must
  	expose a single method with the form:
  			restore()
  			Method sets the collection state for offline access.
  */

	}, {
		key: "getOfflineState",
		value: function getOfflineState() {
			return this[_offlineState];
		}

		/*
  	Experimental feature
  		Sets an object capable of providing a persisted version of this collection. The object must
  	expose a single method with the form:
  			restore()
  			Method sets the collection state for offline access.
  */

	}, {
		key: "setOfflineState",
		value: function setOfflineState(offline) {
			this[_offlineState] = offline;
		}

		//------------------------------------------------------------------------------------------------------
		// Middleware API
		//------------------------------------------------------------------------------------------------------

		/*
  	Registers collection middleware operation. A middleware operation is invoked before each asynchronous/
  	network operation. Middleware operations are functions with the following format:
  		fn(shadow, property)
  		Middleware functions must return a promise and are invoked in the order registered. A function
  	generating an error will terminate the middleware chain and the collection operation will not
  	be performed.
  */

	}, {
		key: "use",
		value: function use(op, mw) {
			var _this2 = this;

			op = Array.isArray(op) ? op : [op];

			(0, _akutils.assert)(function (a) {
				var _iteratorNormalCompletion = true;
				var _didIteratorError = false;
				var _iteratorError = undefined;

				try {
					for (var _iterator = op[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
						var o = _step.value;
						a.has(_this2[_middleware], o);
					}
				} catch (err) {
					_didIteratorError = true;
					_iteratorError = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion && _iterator.return) {
							_iterator.return();
						}
					} finally {
						if (_didIteratorError) {
							throw _iteratorError;
						}
					}
				}
			});

			var _iteratorNormalCompletion2 = true;
			var _didIteratorError2 = false;
			var _iteratorError2 = undefined;

			try {
				for (var _iterator2 = op[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
					var mwop = _step2.value;

					this[_middleware][mwop].push(mw);
				}
			} catch (err) {
				_didIteratorError2 = true;
				_iteratorError2 = err;
			} finally {
				try {
					if (!_iteratorNormalCompletion2 && _iterator2.return) {
						_iterator2.return();
					}
				} finally {
					if (_didIteratorError2) {
						throw _iteratorError2;
					}
				}
			}
		}

		/*
  	Serially invokes each middleware function for the specified operation.
  */

	}, {
		key: "_on",
		value: function _on(op) {
			var _this3 = this;

			var idx = arguments.length <= 1 || arguments[1] === undefined ? 0 : arguments[1];

			(0, _akutils.assert)(function (a) {
				return a.is(_this3[_middleware][op], "Unknown middleware operation: " + op);
			});

			var mw = this[_middleware][op];
			var next = mw[idx];

			if (!next) {
				return _Store2.default.resolve(this._);
			}

			try {
				return next(this._, this, op).then(function () {
					return _this3._on(op, idx + 1);
				});
			} catch (error) {
				return _Store2.default.reject(error);
			}
		}

		/*
  	Experimental feature to update values with any modifications from the server.
  */

	}, {
		key: "resync",
		value: function resync() {
			var values = this._.valuesArray();
			var currSync;
			var lastUpdate = null;
			var initialId = Number.MAX_SAFE_INTEGER;
			var filter, id, v, updatedAt;

			// Todo: break out filter parameters to a pluggable function
			// iterate all values and determine the greatest updated_at
			for (var i = 0, len = values.length; i < len; i++) {
				v = values[i];

				if (!v) {
					continue;
				}

				id = v.id;
				updatedAt = v.updated_at;

				if (id < initialId) {
					initialId = id;
				}

				if (!lastUpdate || lastUpdate < updatedAt) {
					lastUpdate = updatedAt;
				}
			}

			// setup the refresh criteria
			if (lastUpdate) {
				filter = this.endpoint.queryBuilder();

				filter.gte("initial_id", initialId);
				filter.gt("updated_at", lastUpdate);
			}

			debug("resync() - path=" + this.dotPath() + ", initial_id=" + initialId + ", updated_at=" + lastUpdate);

			// Todo: rework network call returns two arrays: new/modified models and deleted model IDs
			//       as currently structured there is no delete support.
			//       What about dirty models? Perhaps they should not be touched if there is a conflict.

			return this.fetch(filter, _CollectionOptions.MERGE_OPTION, !lastUpdate);
		}

		//------------------------------------------------------------------------------------------------------
		// Paging API (could improve by adding two offsets and maintain max pages)
		//------------------------------------------------------------------------------------------------------

	}, {
		key: "fetchNext",
		value: function fetchNext() {
			var _this4 = this;

			var mergeOp = arguments.length <= 0 || arguments[0] === undefined ? _CollectionOptions.MERGE_OPTION : arguments[0];

			if (this.pagingTime) {
				throw new Error("Paging operation in progress");
			} else if (!this.hasMorePages()) {
				return _Store2.default.resolve(this._);
			}

			var filter = this.endpoint.queryBuilder();
			var time = this.pagingTime = Date.now();

			this.set(_paging, true);

			filter.equals("offset", this._[_nextOffset]);
			filter.equals("limit", this._[_limit]);

			return this.fetch(filter, mergeOp, false, function (error, models) {
				// bail if paging times do not match (offset likely reset)
				if (time !== _this4.pagingTime) {
					return;
				}

				_this4.pagingTime = null;
				_this4.set(_paging, false);
				if (error) {
					return;
				}

				_this4.set(_nextOffset, _this4._[_nextOffset] + models.length);
				_this4.set(_lastPageSize, models.length);

				if (models.length < _this4._[_limit]) {
					_this4.set(_synced, true);
				}
			});
		}
	}, {
		key: "isPaging",
		value: function isPaging() {
			var state = arguments.length <= 0 || arguments[0] === undefined ? this._ : arguments[0];

			// use pagingTime instance variable (instant) and _paging state variable (tied to state) to
			// return the most conservative value
			return this.pagingTime || state[_paging];
		}
	}, {
		key: "hasMorePages",
		value: function hasMorePages() {
			var state = arguments.length <= 0 || arguments[0] === undefined ? this._ : arguments[0];

			return this.isConnected() && !state[_synced] && (!state[_lastPageSize] || this._[_lastPageSize] >= this._[_limit]);
		}
	}, {
		key: "nextOffset",
		value: function nextOffset() {
			return this.get(_nextOffset);
		}
	}, {
		key: "resetPaging",
		value: function resetPaging() {
			this.pagingTime = null;
			this.set(_lastPageSize, null);
			this.set(_nextOffset, 0);
			this.set(_paging, false);
		}
	}, {
		key: "setLimit",
		value: function setLimit(limit) {
			this.pagingTime = null;

			this.set(_limit, limit);
			this.set(_lastPageSize, null);
			this.set(_nextOffset, 0);
			this.set(_paging, false);
		}

		//------------------------------------------------------------------------------------------------------
		// Experimental data spec stuff
		//------------------------------------------------------------------------------------------------------

	}, {
		key: "setElementClass",
		value: function setElementClass() {
			var MemberPropertyClass = arguments.length <= 0 || arguments[0] === undefined ? _MapProperty2.default : arguments[0];
			var initialState = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
			var autoShadow = arguments.length <= 2 || arguments[2] === undefined ? true : arguments[2];
			var readonly = arguments.length <= 3 || arguments[3] === undefined ? false : arguments[3];

			var models = this._[_models];
			var modelsShader = models.shader();

			modelsShader.elementShader.setElementClass(MemberPropertyClass, initialState, autoShadow, readonly);
		}

		//------------------------------------------------------------------------------------------------------
		// Endpoint methods
		//------------------------------------------------------------------------------------------------------

	}, {
		key: "clearEndpoint",
		value: function clearEndpoint() {
			this.removeAllModels();
			this.removeProperty(_endpoint);
		}
	}, {
		key: "setEndpoint",
		value: function setEndpoint(endPoint) {
			this.resetPaging();
			this.removeAllModels();
			this.addProperty(_endpoint, endPoint);
		}

		//------------------------------------------------------------------------------------------------------
		// Property subclasses may want to override success methods
		//------------------------------------------------------------------------------------------------------

	}, {
		key: "shadowClass",
		value: function shadowClass() {
			return _CollectionShadow2.default;
		}

		//------------------------------------------------------------------------------------------------------
		// Collection methods
		//------------------------------------------------------------------------------------------------------

	}, {
		key: "addModel",


		/*
  	Synchronously adds a new model object to the store. Call model.$.save() to persist the newly added
  	object.
  		Parmaeters:
  		state - the object model to add to the collection
  		merge - boolean declaring whether this state should be merged over an existing model with
  			the same ID. False means a current model will be replaced with the new model value.
  		Returns the object's ID. And ID is assigned if the 'id' parameter was not set and it could not
  		be found in the 'state' parameter.
  */
		value: function addModel(state) {
			var mergeOp = arguments.length <= 1 || arguments[1] === undefined ? _CollectionOptions.DEFAULTS_OPTION : arguments[1];

			if (!this.isConnected()) {
				throw new Error("Collection is not connected.");
			}

			var id = this.extractId(state);
			var modelId;

			// just add the model
			if (!id || !this.hasModel(id) || mergeOp === _CollectionOptions.REPLACE_OPTION) {
				var modelDefn = _ModelProperty2.default.modelDefinitionFor(state, this);
				var models = this._[_models];

				models.set(modelDefn.cid, modelDefn);

				modelId = modelDefn.id;
			} else {
				var currModel = this._getModel(id);

				switch (mergeOp) {
					case _CollectionOptions.NONE_OPTION:
						break;
					case _CollectionOptions.MERGE_OPTION:
						currModel.merge(state);
						break;
					case _CollectionOptions.REPLACE_OPTION:
						currModel.setData(state);
						break;
					case _CollectionOptions.DEFAULTS_OPTION:
						currModel.defaults(state);
						break;
					default:
						throw new Error("Invalid post-save option: " + mergeOp);
				}

				modelId = currModel.id;
			}

			return modelId;
		}

		/*
  	Bulk adds multiple models. Models must have an ID as it is assumed they have been previously
  	saved.
  		Parameters:
  		models - array of model values
  		merge - boolean declaring whether each state should be merged over an existing model with
  			the same ID. False means a current model will be replaced with the new model value.
  */

	}, {
		key: "addModels",
		value: function addModels(models) {
			var mergeOp = arguments.length <= 1 || arguments[1] === undefined ? _CollectionOptions.DEFAULTS_OPTION : arguments[1];
			var syncOp = arguments.length <= 2 || arguments[2] === undefined ? true : arguments[2];

			if (!this.isConnected()) {
				throw new ("Collection " + this.slashPath + " is not connected")();
			}

			var id, state;

			if (syncOp) {
				this.set(_synced, true);
			}

			for (var i = 0, len = models.length; i < len; i++) {
				this.addModel(models[i], mergeOp);
			}
		}

		/*
  	Combines an add and save actions.
  */

	}, {
		key: "create",
		value: function create(model) {
			if (!this.isConnected()) {
				return _Store2.default.reject("Collection " + this.slashPath + " is not connected");
			}

			var cid = this.addModel(model);

			return this.save(cid);
		}
	}, {
		key: "destroy",
		value: function destroy(id) {
			var _this5 = this;

			if (!this.hasModel(id)) {
				return _Store2.default.resolve(this._);
			}

			try {
				var _ret = function () {
					var model = _this5._getModel(id);

					if (!model || model.isNew()) {
						return {
							v: _Store2.default.resolve(_this5._)
						};
					}

					return {
						v: _this5._on(DestroyOp).then(function () {
							return _this5.endpoint.doDelete(id);
						}).then(function () {
							_this5._[_models].delete(model.cid);
							_this5._[_id2cid].delete(model.id);

							return id;
						}).catch(function (error) {
							return _this5.onError(error, "Destroy model: " + id);
						})
					};
				}();

				if ((typeof _ret === "undefined" ? "undefined" : _typeof(_ret)) === "object") return _ret.v;
			} catch (error) {
				return this.onError(error, "Destroy model: " + id);
			}
		}

		// No reset option - always resets

	}, {
		key: "fetch",
		value: function fetch() {
			var filter = arguments.length <= 0 || arguments[0] === undefined ? null : arguments[0];
			var mergeOp = arguments.length <= 1 || arguments[1] === undefined ? _CollectionOptions.REPLACE_OPTION : arguments[1];

			var _this6 = this;

			var replaceAll = arguments.length <= 2 || arguments[2] === undefined ? true : arguments[2];
			var callback = arguments[3];

			if (!this.isConnected()) {
				return _Store2.default.reject("Collection " + this.slashPath + " is not connected");
			}

			var syncOp = !filter;

			this.setFetching(true);

			try {
				return this._on(FetchOp).then(function () {
					return _this6.endpoint.doFetch(filter);
				}).then(function (models) {
					try {
						_this6.setFetching(false);

						// invoke the callback before processing models
						callback && callback(null, models);

						if (replaceAll) {
							_this6.setModels(models, syncOp);
						} else {
							_this6.addModels(models, mergeOp, syncOp);
						}

						return models;
					} catch (error) {
						return _this6.onError(error, "Fetch error while setting models");
					}
				}).catch(function (error) {
					_this6.setFetching(false);

					// invoke the callback with the error
					callback && callback(error, null);

					return _this6.onError(error, "Fetch all models");
				});
			} catch (error) {
				this.setFetching(false);

				// invoke the callback with the error
				callback && callback(error, null);

				debug("fetch() top level error: " + (error.stack || error));

				throw _Store2.default.reject(error);
			}
		}
	}, {
		key: "find",
		value: function find(id) {
			var _this7 = this;

			if (!this.isConnected()) {
				return _Store2.default.reject("Collection " + this.slashPath + " is not connected");
			}

			try {
				var model = this._getModel(id);

				if (model) {
					return _Store2.default.resolve(model.data);
				} else {
					return this._on(FindOp).then(function () {
						return _this7.endpoint.doFind(id);
					}).then(function (state) {
						_this7.addModel(state, _CollectionOptions.NONE_OPTION);

						return _this7.getModel(id);
					}).catch(function (error) {
						return _this7.onError(error, "Find model " + id);
					});
				}
			} catch (error) {
				return this.onError(error, "Find model " + id);
			}
		}
	}, {
		key: "getModel",
		value: function getModel(id) {
			var state = arguments.length <= 1 || arguments[1] === undefined ? this._ : arguments[1];

			var model = this._getModel(id, state);

			return model && model.data;
		}
	}, {
		key: "hasModel",
		value: function hasModel(id) {
			var state = arguments.length <= 1 || arguments[1] === undefined ? this._ : arguments[1];

			return state[_id2cid].has(id) || state[_models].has(id);
		}

		/*
  	Gets if the collection is active (has a shadow) and an endpoint.
  */

	}, {
		key: "isConnected",
		value: function isConnected() {
			return this._ && this._[_endpoint] && this._[_endpoint].isConnected();
		}
	}, {
		key: "isFetching",
		value: function isFetching() {
			return this[_fetching];
		}
	}, {
		key: "isNew",
		value: function isNew(id) {
			var state = arguments.length <= 1 || arguments[1] === undefined ? this._ : arguments[1];

			var model = this._getModel(id, state);

			return !model || model.isNew();
		}
	}, {
		key: "isNewModel",
		value: function isNewModel(shadow) {
			var id = this.extractId(shadow);

			return !id;
		}

		/*
  	Gets the shadow models currently managed by the collection.
  		Returns an array of models already fetched and/or added.
  */

	}, {
		key: "modelsArray",
		value: function modelsArray(state) {
			if (!state && !this.isConnected()) {
				throw new Error("Collection is not connected.");
			}

			state = state || this._;

			var models = state[_models];
			var keys = Object.keys(models);
			var result = [];

			for (var i = 0, len = keys.length; i < len; i++) {
				result.push(models[keys[i]].data);
			}

			return result;
		}
	}, {
		key: "modelEntries",
		value: function modelEntries(state) {
			var _this8 = this;

			if (!state && !this.isActive()) {
				return _akutils.doneIterator;
			}

			state = state || this._;

			var models = state[_models];
			var keys = this.modelKeysArray(state);

			return (0, _akutils.iterateOver)(keys, function (key) {
				return [key, _this8.getModel(key)];
			});
		}
	}, {
		key: "modelKeys",
		value: function modelKeys() {
			var state = arguments.length <= 0 || arguments[0] === undefined ? this._ : arguments[0];

			// too brute force but quick and sure to work
			return (0, _akutils.iteratorFor)(this.modelKeysArray(state));
		}
	}, {
		key: "modelKeysArray",
		value: function modelKeysArray(state) {
			if (!state && !this.isActive()) {
				return _akutils.doneIterator;
			}

			state = state || this._;

			return Object.keys(state[_models]);
		}
	}, {
		key: "modelValues",
		value: function modelValues() {
			var _this9 = this;

			var state = arguments.length <= 0 || arguments[0] === undefined ? this._ : arguments[0];

			return (0, _akutils.iterateOver)(this.modelKeysArray(state), function (key) {
				return _this9.getModel(key, state);
			});
		}
	}, {
		key: "remove",
		value: function remove(id) {
			if (!this.hasModel(id)) {
				return;
			}

			var model = this._getModel(id);
			var prevShadow = this._;

			this._[_models].delete(model.cid);
			this._[_id2cid].delete(model.id);
		}

		/*
  	Removes all models from the collection and marks the collection as having not synched with the
  	endpoint.
  */

	}, {
		key: "removeAllModels",
		value: function removeAllModels() {
			if (!this.isActive()) {
				return;
			}

			this.set(_synced, false);
			this._[_models].clear();
			this._[_id2cid].clear();
		}

		/*
  	Saves a model through the endpoint.
  		Parameters:
  		id - the model id or cid
  		mergeOp - one of constants: DEFAULTS, MERGE, NONE, REPLACE
  		Returns a promise. The resolve function arguments are F.lux models and this adapter as arguments
  */

	}, {
		key: "save",
		value: function save(id) {
			var _this10 = this;

			var mergeOp = arguments.length <= 1 || arguments[1] === undefined ? _CollectionOptions.MERGE_OPTION : arguments[1];

			if (!this.isConnected()) {
				return _Store2.default.reject("Collection " + this.slashPath + " is not connected");
			}

			var model = this._getModel(id);
			var cid = model && model.cid;
			var shadow = model && model.data;

			if (!model) {
				return _Store2.default.reject("Collection " + this.slashPath + " model not found: id=" + id);
			}

			try {
				var _ret2 = function () {
					var shadowState = shadow.__().nextState();
					var opName = _this10.isNewModel(shadow) ? CreateOp : UpdateOp;

					return {
						v: _this10._on(opName).then(function () {
							if (_this10.isNewModel(shadow)) {
								return _this10.endpoint.doCreate(shadow, shadowState);
							} else {
								return _this10.endpoint.doUpdate(id, shadow, shadowState);
							}
						}).then(function (savedState) {
							var currModel = shadow$(model).latest();
							var savedId = _this10.extractId(savedState);

							// Put an entry in id->cid mapping
							if (savedId != id) {
								_this10._[_id2cid].set(id, cid);
							}

							switch (mergeOp) {
								case _CollectionOptions.NONE_OPTION:
									if (savedId !== id) {
										currModel.changeId(savedId);
									}

									break;
								case _CollectionOptions.MERGE_OPTION:
									currModel.merge(savedState);
									break;
								case _CollectionOptions.REPLACE_OPTION:
									currModel.setData(savedState);
									break;
								case _CollectionOptions.DEFAULTS_OPTION:
									currModel.defaults(savedState);
									break;
								default:
									return _Store2.default.reject("Invalid post-save option: " + mergeOp);
							}

							return _this10.get(id);
						}).catch(function (error) {
							return _this10.onError(error, "Save " + id + " - cid=" + shadow$(shadow).cid);
						})
					};
				}();

				if ((typeof _ret2 === "undefined" ? "undefined" : _typeof(_ret2)) === "object") return _ret2.v;
			} catch (error) {
				return this.onError(error, "Save " + id + " - cid=" + shadow$(shadow).cid);
			}
		}
	}, {
		key: "setIdName",
		value: function setIdName(idName) {
			this._[_idName] = idName;
		}
	}, {
		key: "setFetching",
		value: function setFetching(fetching) {
			this[_fetching] = fetching;
			this.touch();
			//		this.set(_fetching, fetching);
		}

		/*
  	Bulk replaces current models with an array of new models.
  		See comments for addModels().
  */

	}, {
		key: "setModels",
		value: function setModels(models) {
			var syncOp = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

			this.removeAllModels();
			this.addModels(models, _CollectionOptions.REPLACE_OPTION, syncOp);
		}

		//------------------------------------------------------------------------------------------------------
		// Methods Collection subclasses may want to override
		//------------------------------------------------------------------------------------------------------

		/*
  	Generates an ID for a new object. The default implementation generates: temp-[UUID].
  */

	}, {
		key: "makeId",
		value: function makeId() {
			return "temp-" + (0, _akutils.uuid)();
		}

		/*
  	Subclasses can override this method to remove underlying transport specific properties.
  */

	}, {
		key: "clean",
		value: function clean(json) {
			return json;
		}

		/*
  	Used by the private _shadow() method to get the id from the model JSON representation as returned by the
  	subclass doXXX() apis. The default implementation simply returns the 'id' model property.
  */

	}, {
		key: "extractId",
		value: function extractId(model) {
			var idName = this._[_idName];

			return (0, _akutils.isObject)(model) ? model[idName] : model;
		}

		/*
  	Gets an ID suitable for obtaining the shadow model even if the model has yet to be persisted.
  */

	}, {
		key: "extractRefId",
		value: function extractRefId(model) {
			var idName = this._[_idName];

			if (!(0, _akutils.isObject)(model)) {
				return model;
			}

			if (model[idName]) {
				return model[idName];
			} else if (model.$) {
				return shadow$(model).$$.cid;
			}
		}
	}, {
		key: "onError",
		value: function onError(error, opMsg) {
			var msg;

			if (error.status) {
				var statusText = error.status ? error.response.statusText || "HTTP Code=" + error.status : "HTTP status code not specified";
				msg = "Error during collection operation '" + opMsg + " (" + this.endpointId + ")' - " + ("Server Error: status=" + statusText);
			} else if (error.message) {
				msg = "Error during collection operation '" + opMsg + " (" + this.endpointId + ")' - Collection Error: " + error.message;
			} else {
				msg = "Error during collection operation '" + opMsg + " (" + this.endpointId + ")' - Error: " + error;
			}

			debug(_debug.CollectionPropertyKey + " Error: " + msg);
			if (error.stack) {
				debug(error.stack);
			}

			var collectionError = new Error(msg);

			collectionError.status = error.status;
			collectionError.endpointError = error;

			this.emit(ErrorEvent, collectionError, this._, this);

			return _Store2.default.reject(collectionError);
		}

		//------------------------------------------------------------------------------------------------------
		// Private methods
		//------------------------------------------------------------------------------------------------------

		/*
  	Gets the Model container object NOT the actual model.
  */

	}, {
		key: "_getModel",
		value: function _getModel(id) {
			var state = arguments.length <= 1 || arguments[1] === undefined ? this._ : arguments[1];

			var id2cid = state[_id2cid];
			var cid = id2cid.has(id) ? id2cid.get(id) : id;

			return state[_models].get(cid);
		}
	}, {
		key: "endpoint",
		get: function get() {
			return this.isActive() && this._[_endpoint];
		}
	}, {
		key: "endpointId",
		get: function get() {
			var endpoint = this.endpoint;

			return endpoint && endpoint.id;
		}
	}, {
		key: "modelsCount",
		get: function get() {
			return this._[_models].size;
		}
	}]);

	return CollectionProperty;
}(_KeyedProperty3.default);

// Mix in `Emitter`


exports.default = CollectionProperty;
(0, _componentEmitter2.default)(CollectionProperty.prototype);